/**
 * 
 */
package org.jfree.eastwood;

import java.awt.AlphaComposite;
import java.awt.Composite;
import java.awt.Font;
import java.awt.GradientPaint;
import java.awt.Graphics2D;
import java.awt.Paint;
import java.awt.Shape;
import java.awt.Stroke;
import java.awt.geom.Point2D;
import java.awt.geom.Rectangle2D;
import java.util.Iterator;

import org.jfree.chart.ChartRenderingInfo;
import org.jfree.chart.JFreeChart;
import org.jfree.chart.entity.EntityCollection;
import org.jfree.chart.entity.JFreeChartEntity;
import org.jfree.chart.event.ChartProgressEvent;
import org.jfree.chart.plot.Plot;
import org.jfree.chart.plot.PlotRenderingInfo;
import org.jfree.chart.title.Title;
import org.jfree.ui.Align;

/**
 *
 * @author Servoy Stuff
 * http://www.servoy-stuff.net/
 *
 */
public class ExtendedJFreeChart extends JFreeChart {

    /** The angle in radians. */
    private double angle;

    /** The factor for the start of the background gradient. */
    private double f0;

    /** The factor for the end of the background gradient. */
    private double f1;

    /**
	 * Constructor
	 * @param plot
	 */
	public ExtendedJFreeChart(Plot plot) {
		super(plot);
	}

	/**
	 * Constructor
	 * @param title
	 * @param titleFont
	 * @param plot
	 * @param createLegend
	 */
	public ExtendedJFreeChart(String title, Font titleFont, Plot plot,
			boolean createLegend) {
		super(title, titleFont, plot, createLegend);
	}

	/**
	 * Constructor
	 * @param title
	 * @param plot
	 */
	public ExtendedJFreeChart(String title, Plot plot) {
		super(title, plot);
	}
	
    /**
     * Sets the angle (in radians).
     *
     * @param angle  the angle.
     */
    public void setAngle(double angle) {
        this.angle = angle;
    }

    /**
     * Sets the factor for the beginning of the background gradient.
     *
     * @param f0  the factor.
     */
    public void setF0(double f0) {
        this.f0 = f0;
    }

    /**
     * Sets the factor for the end of the background gradient.
     *
     * @param f1  the factor.
     */
    public void setF1(double f1) {
        this.f1 = f1;
    }

    /**
     * Draws the chart on a Java 2D graphics device (such as the screen or a
     * printer).
     * <P>
     * This method is the focus of the entire JFreeChart library.
     *
     * @param g2  the graphics device.
     * @param chartArea  the area within which the chart should be drawn.
     * @param anchor  the anchor point (in Java2D space) for the chart
     *                (<code>null</code> permitted).
     * @param info  records info about the drawing (null means collect no info).
     */
    public void draw(Graphics2D g2,
                     Rectangle2D chartArea, Point2D anchor,
                     ChartRenderingInfo info) {

        notifyListeners(new ChartProgressEvent(this, this,
                ChartProgressEvent.DRAWING_STARTED, 0));
        
        EntityCollection entities = null;
        // record the chart area, if info is requested...
        if (info != null) {
            info.clear();
            info.setChartArea(chartArea);
            entities = info.getEntityCollection();
        }
        if (entities != null) {
        	entities.add(new JFreeChartEntity((Rectangle2D) chartArea.clone(),
                    this));
        }

        // ensure no drawing occurs outside chart area...
        Shape savedClip = g2.getClip();
        g2.clip(chartArea);

        g2.addRenderingHints(getRenderingHints());

        // draw the chart background...
        if (getBackgroundPaint() != null) {
        	drawBackground(g2,chartArea);
            //g2.setPaint(getBackgroundPaint());
            //g2.fill(chartArea);
        }

        if (getBackgroundImage() != null) {
            Composite originalComposite = g2.getComposite();
            g2.setComposite(AlphaComposite.getInstance(AlphaComposite.SRC_OVER,
                    getBackgroundImageAlpha()));
            Rectangle2D dest = new Rectangle2D.Double(0.0, 0.0,
            		getBackgroundImage().getWidth(null),
            		getBackgroundImage().getHeight(null));
            Align.align(dest, chartArea, getBackgroundImageAlignment());
            g2.drawImage(getBackgroundImage(), (int) dest.getX(),
                    (int) dest.getY(), (int) dest.getWidth(),
                    (int) dest.getHeight(), null);
            g2.setComposite(originalComposite);
        }

        if (isBorderVisible()) {
            Paint paint = getBorderPaint();
            Stroke stroke = getBorderStroke();
            if (paint != null && stroke != null) {
                Rectangle2D borderArea = new Rectangle2D.Double(
                        chartArea.getX(), chartArea.getY(),
                        chartArea.getWidth() - 1.0, chartArea.getHeight()
                        - 1.0);
                g2.setPaint(paint);
                g2.setStroke(stroke);
                g2.draw(borderArea);
            }
        }

        // draw the title and subtitles...
        Rectangle2D nonTitleArea = new Rectangle2D.Double();
        nonTitleArea.setRect(chartArea);
        getPadding().trim(nonTitleArea);

        if (getTitle() != null) {
            EntityCollection e = drawTitle(getTitle(), g2, nonTitleArea,
                    (entities != null));
            if (e != null) {
                entities.addAll(e);
            }
        }

        Iterator iterator = getSubtitles().iterator();
        while (iterator.hasNext()) {
            Title currentTitle = (Title) iterator.next();
            if (currentTitle.isVisible()) {
                EntityCollection e = drawTitle(currentTitle, g2, nonTitleArea,
                        (entities != null));
                if (e != null) {
                    entities.addAll(e);
                }
            }
        }

        Rectangle2D plotArea = nonTitleArea;

        // draw the plot (axes and data visualisation)
        PlotRenderingInfo plotInfo = null;
        if (info != null) {
            plotInfo = info.getPlotInfo();
        }
        getPlot().draw(g2, plotArea, anchor, null, plotInfo);

        g2.setClip(savedClip);

        notifyListeners(new ChartProgressEvent(this, this,
                ChartProgressEvent.DRAWING_FINISHED, 100));
    }

    /**
     * Draws the background for the plot.
     *
     * @param g2  the graphics device.
     * @param area  the area.
     */
    public void drawBackground(Graphics2D g2, Rectangle2D area) {
        Paint p = getBackgroundPaint();
        if (p instanceof GradientPaint) {
            GradientPaint gp = (GradientPaint) p;
            double x0 = area.getMinX();
            double x1 = x0;
            double y0 = area.getMaxY();
            double y1 = y0;
            double w = area.getWidth();
            double h = area.getHeight();
            if (this.angle <= 0.0) {
                // horizontal gradient
                x0 = x0 + this.f0 * w;
                x1 = x1 + this.f1 * w;
            }
            else if (this.angle >= Math.PI / 2.0) {
                // vertical gradient
                y0 = y0 - this.f0 * h;
                y1 = y1 - this.f1 * h;
            }
            else {
                // general case
                double r = Math.sqrt(w * w + h * h);
                x0 = x0 + Math.cos(this.angle) * r * this.f0;
                y0 = y0 - Math.sin(this.angle) * r * this.f0;
                x1 = x1 + Math.cos(this.angle) * r * this.f1;
                y1 = y1 - Math.sin(this.angle) * r * this.f1;
            }

            // create the new gradient paint
            Point2D p0 = new Point2D.Double(x0, y0);
            Point2D p1 = new Point2D.Double(x1, y1);
            p = new GradientPaint(p0, gp.getColor1(), p1,
                    gp.getColor2());
        }
        g2.setPaint(p);
        g2.fill(area);
    }

}
